#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <time.h>


#define SF_SIZEOF(a) (a == SF_FLOAT ? sizeof(t_float) : \
                     a == SF_16BIT ? sizeof(short) : 1)

typedef unsigned char uchar;

typedef struct _glob
{
    int sendfd;
    int recfd;
    int count;
    int nflag;
} t_glob;

typedef struct _tag
{                           /* size (bytes) */
    char tag[4];            /*  4  */ /*"TAG!"*/
    char format;            /*  1  */
    long count;             /*  4  */
    char channels;          /*  1  */
    long framesize;         /*  4  */
    char reserved[2];       /*  2  */ /* pad to 16 bytes */
} t_tag;                    /*-----*/
                            /* 16  */

pthread_mutex_t x_mutex = PTHREAD_MUTEX_INITIALIZER;

int recconn(t_glob *gb) {

  struct sockaddr_in server;
  int err, portno = 8001;


    /* create a socket */
    if ((gb->recfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
         err = errno;
         printf("Error! socket: %s\n", strerror(err));
         return 1;
    }

    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;

    /* assign server port number */
    server.sin_port = htons((u_short)portno);
    printf(" Listening to port number %d for sync\n", portno);

    /* name the socket */
    if (bind(gb->recfd, (struct sockaddr *)&server, sizeof(server)) < 0)
    {
         err = errno;
         printf("Error! bind: %s\n", strerror(err));;
         return 2;
    }
    return 0;
}

int sendconn(t_glob *gb) {

  char *hostname = "localhost";
  int portno, err;
  struct sockaddr_in  server;
  struct hostent *hp;

    hostname = "localhost";
    portno = 8002;
    gb->count = 0;

    /* create a socket */
    gb->sendfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (gb->sendfd < 0)
    {
         printf(" Connection to %s on port %d failed", hostname, portno); 
	 err = errno;
	 printf("Error! socket: %s\n", strerror(err));
         return (0);
    }

    /* connect socket using hostname provided in command line */
    server.sin_family = AF_INET;
    hp = gethostbyname(hostname);
    if (hp == 0)
    {
        printf("Error! bad host?");
        return 1;
    }

    memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length);

    /* assign client port number */
    server.sin_port = htons((unsigned short)portno);

    /* try to connect */
    if (connect(gb->sendfd, (struct sockaddr *) &server, sizeof (server)) < 0)
    {
        err = errno;
        printf("Error! connect: %s\n", strerror(err));
        return 1;
    }

    printf(" Connected stream to %s on port %d\n", hostname, portno);

    pthread_mutex_lock(&x_mutex);
    pthread_mutex_unlock(&x_mutex);
    return (0);
}

int dodata(t_glob *gb){

  int blocksize = 2048;
  int length = blocksize * sizeof(short);
  int err, i;
  t_tag x_tag;
  char buf[length];
  short *bufptr;
  ssize_t bytes;
  
        if ( (bytes = read(0, &buf, length)) != length ) {
		printf("Finished reading stdin - block count: %i\n", gb->count);
		return 1;
	}

	if(!gb->nflag)
	{
	  i = blocksize; bufptr =  (short*) &buf;
	  while(i--)
		*bufptr++ = htons(*bufptr);
	}

	gb->count++;
        if (gb->sendfd != -1)
        {
            //bp = (char *)x->x_cbuf;
            /* fill in the header tag */
            x_tag.tag[0] = 'T';
            x_tag.tag[1] = 'A';
            x_tag.tag[2] = 'G';
            x_tag.tag[3] = '!';
	    x_tag.format = 11;	/* SF_16BIT */
            x_tag.count = htonl(gb->count);
	    x_tag.channels = 1;
            x_tag.framesize = htonl(length);

            /* send the format tag */
            if (send(gb->sendfd, (char*)&x_tag, sizeof(t_tag), 0) < 0)
            {
		err = errno;
		printf("udpsend~: send tag: %s (%d)\n", strerror(err), err);
		return 1;
            }
                /* If there is any data, send the buffer, the OS might segment it into smaller packets */
            if (send(gb->sendfd, &buf, length, 0) <= 0) {
		err = errno;
		printf("udpsend~: send data: %s (%d)\n", strerror(err), err);
		return 1;
            }
	}
	//nanosleep(&tm, NULL);
	return 0;
}

void usage(char **argv){

	printf("Usage:\n\t '16bit_mono_raw_out' | %s -h (or -n)\n", argv[0]);
	printf("\t\t-h for 'host-byte order'\n\t\t-n for 'network byte order\n");
	printf("\neg. mpg123 -m -s song.mp3 | %s -h\n\n");
}

int main(int argc, char **argv){

  int ret;
  ssize_t bytes;
  char buf;
  t_glob x_gb;

	x_gb.nflag = 0;

	if (argc<2){
	   usage(argv);
	   return 1;
	}
	while ((ret = getopt (argc, argv, "hn")) != -1)
          switch (ret)
          {
           case 'h':
             break;
	   case 'n':
             x_gb.nflag = 1;
	     break;
	   default:
	     usage(argv);
	     return 1;
	  }


	if ( ((ret = recconn(&x_gb)) != 0) || ((ret = sendconn(&x_gb)) != 0) ) {
		return 1;
	}

        if ((ret = dodata(&x_gb)) != 0) {
                return 1;
        }

    while (1){
	bytes = read(x_gb.recfd, &buf, 1);
	if ((ret = dodata(&x_gb)) != 0) {
		break;
	}
    }
        return 0;

}
